/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Quickfile extension for Thunderbird.
 *
 * The Initial Developer of the Original Code is Paul Tomlin.
 * Portions created by the Initial Developer are Copyright (C) 2005, 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Paul Tomlin
 *  Mark Shuttleworth
 *
 * ***** END LICENSE BLOCK ***** */

/* $Id: quickfile.js,v 1.16 2006/11/20 10:59:57 paultomlin Exp $ */

const QF_INSTACOPY  = 'instacopy';
const QF_INSTAMOVE  = 'instamove';
const QF_COPY       = 'copy';
const QF_MOVE       = 'move';
const QF_NAVIGATE   = 'navigate';

var Quickfile =
{
  /**************************************************************
  ** PUBLIC interface
  **************************************************************/

  /**
   * Called by the XUL dialog upon loading
   */
  onLoad: function _onLoad ()
  {
    this._log("Quickfile::onLoad()");

    this._loadPreferences();

    this._initMode(window.arguments[0]);
    if (! this._mode) {
      this._log("  invalid mode");
      window.close();
      return;
    }

    var textbox = document.getElementById('textbox');

    // Set the search root for the autocomplete.
    // This is doubled up due to support of XPFE and Toolkit
    // XPFE is handled by setting an nsICategoryManager entry on the autocomplete session
    // Toolkit is handled by setting the autocompletesearchparams attribute
    var autoCompleteSession = textbox.getSession(0);
    if (autoCompleteSession && typeof autoCompleteSession.QueryInterface == 'function') {
      autoCompleteSession.QueryInterface(nsICategoryManager)
                         .addCategoryEntry('quickfile', 'searchRoot', this._searchRootUri, false, true);
    }
    textbox.setAttribute('autocompletesearchparam', this._searchRootUri);

    // If we know about a default folder URI, then check it is actually valid
    // before remembering the display name and setting the textfield value
    if (this._defaultFolderUri) {
      var folder = QuickfileCommon.getFolderFromURI(this._defaultFolderUri);
      if (folder) {
        // Get the default folder name, relative to our search root, if it's relavent
        var defaultFolderName = QuickfileCommon.getFolderRelativeName(folder, this._searchRootFolder);
        this._defaultFolderName = defaultFolderName;
        textbox.value = defaultFolderName;
      }
      else
        this._defaultFolderUri = null;
    }

    // TODO: find a cleaner way of doing this
    // inelligent "hack" described by Neil to get the popup items
    // cropped on the left rather than the right. i don't know
    // how or why it works
    document.getElementById('value').setAttribute('crop', 'start');

    if (this._mode == QF_INSTACOPY || this._mode == QF_INSTAMOVE)
      this.onDialogAccept();
  },

  /**
   * Called by the autocomplete textbox when the user types something
   */
  onTextInput: function _onTextInput ()
  {
    this._log("Quickfile::onTextInput()");

    this._destFolder = null;

    return true;
  },

  /**
   * Called by the autocomplete textbox when the user selects a result
   */
  onTextEntered: function _onTextEntered ()
  {
    this._log("Quickfile::onTextEntered()");

    //document.documentElement.acceptDialog();

    return true;
  },

  /**
   * Called by the autocomplete textbox when the user cancels editing
   */
  onTextReverted: function _onTextReverted ()
  {
    this._log("Quickfile::onTextReverted()");

    this._destFolder = null;

    return true;
  },

  /**
   * Called by the XUL dialog when the user accepts the dialog
   */
  onDialogAccept: function _onDialogAccept ()
  {
    this._log("Quickfile::onDialogAccept()");

    this._selectedName = document.getElementById('textbox').value;

    // trim leading/trailing slashes
    this._selectedName = this._selectedName.replace(/^\/*/, '').replace(/\/*$/, ''); /* eclipse */

    // Our binding on autocomplete adds a 'uriValue' property, stolen from
    // the comment property of the individual result. A bit of a poor mans
    // name/value lookup but seems to work well enough to use.
    this._selectedUri = document.getElementById('textbox').uriValue;

    // When a default folder is employed, and the user hasn't changed
    // anything in the textbox, no autocomplete is done, so the text
    // we get back here is exactly what we put there during onLoad.
    if (! this._selectedUri && this._selectedName == this._defaultFolderName)
      this._selectedUri = this._defaultFolderUri;

    // When a custom search root is in use, the text displayed, and hence returned
    // from the textbox.value property, doesn't include the path to the search
    // root. We need to add that prefix now
    if (this._searchRootUri != QF_SEARCHROOT_ALL && this._searchRootUri != QF_SEARCHROOT_ACCOUNT) {
      var searchRoot = QuickfileCommon.getFolderFromURI(this._searchRootUri);

      // prepend the folder with any prefix resulting from the search root
      this._selectedName = QuickfileCommon.getFolderRelativeName(searchRoot, null) + '/' + this._selectedName;
    }

    // Get a folder and check the URI is actually valid
    this._destFolder = QuickfileCommon.getFolderFromURI(this._selectedUri);

    // if the destination folder doesn't exist, but the user has mkdir enabled
    // go off and create the folder. since imap is handled async, we return false
    // here to keep the dialog open, and wait for the _onDestFolderReady call
    if (! this._destFolder && this._mkdir) {
      this._destFolder = this._createFolderForName(this._selectedName);
      return false;
    }

    this._onDestFolderReady();

    return false;
  },

  /**************************************************************
  ** PRIVATE data
  **************************************************************/

  /**
   * Output debugging info or not
   */
  _debug: true,

  /**
   * Where the user wants to search from. Default to everything, load
   * user settings from prefs
   */
  _searchRootUri: 'msgaccounts:/',

   /**
    * Unless the search root is "all accounts" [msgaccounts:/] then we
    * can use create an nsIMsgFolder for it, for relative naming purposes
    */
  _searchRootFolder: null,

  /**
   * The window that opened us
   */
  _window: window.opener,

  /**
   * The nsIMsgFolder to copy/move the message[s] to
   */
  _destFolder: null,

  /**
   * The mode of operation
   */
  _mode: null,

  /**
   * The default folder URI to use, if any
   */
  //   mailbox://nobody@Local%20Folders/Inbox
  _defaultFolderUri: null,

  /**
   * The default folder, nsIMsgFolder
   */
  _defaultFolder: null,

  /**
   * The display name for the default folder
   * This is used to check if the user hasn't edited the textbox
   * since the autocomplete doesn't do a lookup in that instance
   */
  _defaultFolderName: null,

  /**
   * Do we create folders that the user selects but which don't
   * already exist?
   */
  _mkdir: true,

  /**
   * During folder creation, we store the array of folders
   * awaiting creation here
   */
  _createFolderSequence: null,

  /**
   * The URI returned from the textbox, if any
   */
  _selectedUri: null,

  /**
   * The display name from the textbox
   */
  _selectedName: null,

  /**************************************************************
  ** PRIVATE methods
  **************************************************************/

  /**
   *
   */
  _log: QuickfileCommon.log,

  /**
   *
   */
  _loadPreferences: function __loadPreferences ()
  {
    this._log("Quickfile::_loadPreferences()");

    var prefs = Components.classes["@mozilla.org/preferences-service;1"]
            .getService(nsIPrefService)
            .getBranch("quickfile.");

    if (prefs.getPrefType('debug.enable') == prefs.PREF_BOOL)
      this._debug = prefs.getBoolPref('debug.enable');

    if (prefs.getPrefType('search.root') == prefs.PREF_STRING) {
      this._searchRootUri = prefs.getCharPref('search.root');
      if (this._searchRootUri != QF_SEARCHROOT_ALL)
        this._searchRootFolder = QuickfileCommon.getFolderFromURI(this._searchRootUri);
    }

    if (prefs.getPrefType('defaultfolder.enable') == prefs.PREF_BOOL && prefs.getBoolPref('defaultfolder.enable')) {
      if (prefs.getPrefType('defaultfolder.uri') == prefs.PREF_STRING) {
        this._defaultFolderUri = prefs.getCharPref('defaultfolder.uri');
        this._defaultFolder = QuickfileCommon.getFolderFromURI(this._defaultFolderUri);
      }
    }

    if (prefs.getPrefType('mkdir.enable') == prefs.PREF_BOOL)
      this._mkdir = prefs.getBoolPref('mkdir.enable');
  },

  /**
   *
   */
  _initMode: function __initMode (aMode)
  {
    this._log("Quickfile::_initMode(" + aMode + ")");

    switch (aMode) {
      case QF_INSTACOPY:
      case QF_INSTAMOVE:
      case QF_COPY:
      case QF_MOVE:
      case QF_NAVIGATE:
        this._mode = aMode;
        document.title += ' - [' + aMode + ']';
        break;
      default:
        this._mode = null;
        break;
    }
  },

  /**
   * Get an array of the URIs of the selected messages
   *
   * @return Array of messages currently selected
   */
  _getSelectedMessages: function __getSelectedMessages ()
  {
    this._log("Quickfile::_getSelectedMessages()");

    return this._window.GetSelectedMessages ? this._window.GetSelectedMessages() : [];
  },

  /**
   *
   */
  _onDestFolderReady: function __onDestFolderReady ()
  {
    this._log("Quickfile::_onDestFolderReady()");

    if (! this._destFolder) {
      this._log("  _destFolder not valid");
      this._onDestFolderInvalid();
      return false;
    }

    // Try to ensure that the uriValue and value of the textbox were
    // in sync. If forceComplete is not enabled, for instance, then
    // the getOverrideValue of the autocomplete-result-popup binding has
    // different rules than the default autocomplete.
    // Ensure we're dealing with absolute names here, as this is really the
    // last chance to pick up an error from that
    if (this._selectedName != QuickfileCommon.getFolderRelativeName(this._destFolder, null)) {
      this._log("  selected text and destination folder name mismatch");
      this._destFolder = null;
      this._onDestFolderInvalid();
      return false;
    }

    switch (this._mode) {
      case QF_INSTAMOVE:
      case QF_MOVE:
        this._transferMessages(true);
        return false;  // leave dialog open during transfer
      case QF_INSTACOPY:
      case QF_COPY:
        this._transferMessages(false);
        return false;  // leave dialog open during transfer
      case QF_NAVIGATE:
        this._selectFolder();
        window.close();
        return true;  // close dialog
    }

    // invalid mode?
    return false;
  },

  /**
   *
   */
  _onDestFolderInvalid: function __onDestFolderInvalid ()
  {
    this._log("Quickfile::_onDestFolderInvalid()");

    this._showProgress(false);

    alert("destination folder invalid");
  },

  /**
   * Transfer the selected messages to the destination
   *
   * @param aIsMove is the transfer a move operation
   *                if true, the messages are moved
   *                otherwise they're copied
   */
  _transferMessages: function __transferMessages (aIsMove)
  {
    this._log("Quickfile::_transferMessages(" + aIsMove + ")");

    var msgs = this._getSelectedMessages();
    if (! (msgs || msgs.length))
      return;

    // build an nsISupportsArray of nsIMsgDBHdrs
    var msgDbHdrs = Components.classes['@mozilla.org/supports-array;1']
                              .createInstance(nsISupportsArray);

    for (var i = 0; i < msgs.length; i++) {
      var msgDBHdr = this._getMsgDbHdrFromURI(msgs[i]);
      if (msgDBHdr)
        msgDbHdrs.AppendElement(msgDBHdr);
    }

    // NOTE: We presume here that all the selected messages are in the same
    //       folder. This will break for some complex views (which I've never
    //       seen an example of).
    var srcFolder = msgDbHdrs.GetElementAt(0).folder;
    if (! srcFolder || srcFolder == this._destFolder)
      return;

    var msgCopyService = Components.classes['@mozilla.org/messenger/messagecopyservice;1']
                                   .getService(nsIMsgCopyService);
    if (! msgCopyService)
      return;

    var msgMailSession = Components.classes['@mozilla.org/messenger/services/session;1']
                                   .getService(nsIMsgMailSession);
    if (! msgMailSession)
      return;

    // now we are about to do the transfer, show the progress bar and info
    this._setLabel('Transferring messages');
    this._showProgress(true);

    msgCopyService.CopyMessages(srcFolder, msgDbHdrs, this._destFolder, aIsMove, this, msgMailSession.topmostMsgWindow, true);
  },

  /**
   * Called by the nsIMsgCopyServiceListener implementation below
   * when the transfer is complete.
   *
   * @param aStatus  the status of the copy operation.
   *                 0 appears to be success
   */
  _onTransferComplete: function __onTransferComplete (aStatus)
  {
    this._log("Quickfile::_onTransferComplete(" + aStatus + ")");

    window.close();
  },

  /**
   * Navigate to (select) the destination folder
   */
  _selectFolder: function __selectFolder ()
  {
    this._log("Quickfile::_selectFolder()");
    if (! this._destFolder)
      return false;

    // Cheap escape, SelectFolder is a JS function defined on mail windows
    // If it exists, try to use it.
    // NOTE: I have a bit of a philisophical issue with this method, since it's a private
    //       API of the mail windows. The nsI* are published, documented interfaces, so I
    //       feel happier about using them. Neil didn't agree, so this is here while I
    //       procrastinate about the decision of which code to use.

    if (this._window && this._window.SelectFolder && typeof this._window.SelectFolder == 'function') {
      this._window.SelectFolder(this._destFolder.URI);
      return true;
    }

    var msgMailSession = Components.classes['@mozilla.org/messenger/services/session;1']
                                   .getService(nsIMsgMailSession);
    if (! msgMailSession)
      return false;

    var topWindow = msgMailSession.topmostMsgWindow;
    if (! topWindow)
      return false;

    // SEE: bug #345251
    //      There was a change on trunk as part of some clean up. mscott moved the
    //      SelectFolder method to windowCommands, and in the process changed the
    //      name to selectFolder.
    var doneSelectFolder = false;
    // assume TB<=2
    try {
      topWindow.SelectFolder(this._destFolder.URI);
      doneSelectFolder = true;
    }
    catch (ex) {
      this._log("  failed selcting folder via topWindow.SelectFolder: " + ex);
      doneSelectFolder = false;
    }

    // otherwise try TB>2
    if (! doneSelectFolder) {
      try {
        topWindow.windowCommands.selectFolder(this._destFolder.URI);
        doneSelectFolder = true;
      }
      catch (ex) {
        this._log("  failed selcting folder via topWindow.windowCommands.selectFolder: " + ex);
        doneSelectFolder = false;
      }
    }

    if (! doneSelectFolder)
      return false;

    // This idea is to select the most relevent message from the folder we've
    // just selected, be that the first unread, or whatever.
    // TODO: This is using TB code that we've got no control over. I can't
    //       find a way to do this via XPCOM, yet.
    if (this._window.GoNextMessage && typeof this._window.GoNextMessage == 'function')
      this._window.GoNextMessage(nsMsgNavigationType.nextMessage, false);

    return true;
  },

  /**
   *
   */
  _showProgress: function __showProgress (aProgressVisible)
  {
    document.getElementById('textbox').hidden = aProgressVisible;
    document.getElementById('progress').hidden = ! aProgressVisible;
    document.documentElement.getButton('accept').disabled = aProgressVisible;
    document.documentElement.getButton('cancel').disabled = aProgressVisible;
    document.documentElement.getButton('help').disabled = aProgressVisible;

    window.setCursor(aProgressVisible ? 'wait' : '');
  },

  /**
   *
   */
  _setLabel: function __setLabel (aLabel)
  {
    document.getElementById('label').value = aLabel;
  },

  /**
   * Get an nsIMsgDBHdr from a URI
   *
   * @param aUri the URI of the message you seek
   * @return nsIMsgDBHdr
   */
  _getMsgDbHdrFromURI: function __getMsgHdrFromURI (aUri)
  {
    return Components.classes['@mozilla.org/messenger;1']
                     .createInstance(nsIMessenger)
                     .messageServiceFromURI(aUri)
                     .messageURIToMsgHdr(aUri);
  },

  /**
   *
   */
  _createFolderForName: function __createFolderForName (aFolderName)
  {
    this._log("Quickfile::_createFolderForName(" + aFolderName + ")");

    var segments = aFolderName.split('/');

    // find the root folder the new folder is going to be under
    var rootFolderName = segments.shift();
    var folder = this._getRootFolderForName(rootFolderName);
    this._log(" found root folder: " + folder.URI);

    // we know imap is going to take some time, show the progress
    if (folder.URI.substr(0, 5) == 'imap:') {
      this._setLabel('Creating folders');
      this._showProgress(true);
    }

    this._log(" finding existing segments");
    for (var i = 0; i < segments.length; i++) {
      var segment = segments[0];
      var f = this._getSubfolder(folder, segment);
      if (f) {
        segments.shift();
        folder = f;
      }
      else
        break;
    }

    // folder now holds the closest existing folder in the account
    // the rest need to be made up from the user entered text
    this._log(" segments left: " + segments.join('/'));

    this._createFolderSequence = [];
    for (var i = 0; i < segments.length; i++) {
      this._createFolderSequence.push(folder.URI + '/' + segments.slice(0, i + 1).join('/'));
    }
    this._log(" this._createFolderSequence\n" + this._createFolderSequence.join("\n"));

    this._continueCreateFolderForName();
  },

  /**
   *
   */
  _continueCreateFolderForName: function __continueCreateFolderForName ()
  {
    this._log("Quickfile::_continueCreateFolderForName(...)");

    if (this._createFolderSequence.length) {
      var uri = this._createFolderSequence.shift();
      this._destFolder = QuickfileCommon.getFolderFromURI(uri);
      try {
        this._destFolder.createStorageIfMissing(this);
      }
      catch (ex) {
        this._onDestFolderInvalid();
      }

      // imap is async, via the nsIUrlListener interface
      // everything else just gets done syncronously
      if (uri.substr(0, 5) != 'imap:')
        this._continueCreateFolderForName();
    }
    else
      this._onDestFolderReady();
  },

  /**
   *
   */
  _getSubfolder: function __getSubfolder (aFolder, aSubfolderName)
  {
    this._log("Quickfile::_getSubfolder(" + aFolder.URI + ", " + aSubfolderName + ")");

    var folder = null;

    if (aFolder.hasSubFolders) {
      var subFoldersEnumerator = aFolder.GetSubFolders();
      try {
        subFoldersEnumerator.first();
        do {
          var subFolder = subFoldersEnumerator.currentItem().QueryInterface(nsIMsgFolder);
          if (subFolder.name == aSubfolderName) {
            folder = subFolder;
            break;
          }

          subFoldersEnumerator.next();
        }
        while (Components.lastResult == 0);
      }
      catch (ex) {
        // nsIEnumerator is a broken interface,
        // it actually throws an exception when next() is at the end
        //this.log("nsIEnumerator is broken! [" + ex + "]");
      }
    }

    return folder;
  },

  /**
   *
   */
  _getRootFolderForName: function __getRootFolderForName (aRootFolderName)
  {
    this._log("Quickfile::_getRootFolderForName(" + aRootFolderName + ")");

    // find the account the folder is going to be under
    var accounts = Components.classes['@mozilla.org/messenger/account-manager;1']
                             .createInstance(nsIMsgAccountManager)
                             .accounts
                             .QueryInterface(nsISupportsArray);

    for (var i = 0; i < accounts.Count(); i++) {
      var account = accounts.GetElementAt(i).QueryInterface(nsIMsgAccount);
      var rootFolder = account.incomingServer.rootMsgFolder;
      if (rootFolder.name == aRootFolderName)
        return rootFolder;
    }

    return null;
  },

  /**************************************************************
  ** nsIMsgCopyServiceListener interface
  **************************************************************/

  /**
   * @see bug #263997 for why this isn't called
   */
  OnProgress: function _OnProgress (aProgress, aProgressMax)
  {
    this._log('Quickfile::msgCopyListener::OnProgress(' + aProgress + ', ' + aProgressMax + ')');
  },

  /**
   *
   */
  OnStartCopy: function _OnStartCopy ()
  {
    this._log('Quickfile::msgCopyListener::OnStartCopy()');
  },

  /**
   *
   */
  OnStopCopy: function _OnStopCopy (aStatus)
  {
    this._log('Quickfile::msgCopyListener::OnStopCopy(' + aStatus + ')');

    this._onTransferComplete(aStatus);
  },

  /**
   *
   */
  SetMessageKey: function _SetMessageKey (aKey)
  {
    this._log('Quickfile::msgCopyListener::SetMessageKey(' + aKey + ')');
  },


  /**************************************************************
  ** nsIUrlListener interface
  **************************************************************/

  /**
   *
   */
  OnStartRunningUrl: function _OnStartRunningUrl (aUrl)
  {
    this._log('Quickfile::urlListener::OnStartRunningUrl(' + aUrl.asciiSpec + ')');
  },

  /**
   *
   */
  OnStopRunningUrl: function _OnStopRunningUrl (aUrl, aExitCode)
  {
    this._log('Quickfile::urlListener::OnStopRunningUrl(' + aUrl.asciiSpec + ', ' + aExitCode + ')');

    // TODO: check exit code
    this._continueCreateFolderForName();
  },

  /**************************************************************
  ** nsISupports interface
  **************************************************************/

  /**
   *
   */
  QueryInterface: function _QueryInterface (aIID)
  {
    if (!aIID.equals(nsIMsgCopyService) && !aIID.equals(nsISupports))
      throw NS_ERROR_NO_INTERFACE;

    return this;
  }
};

window.addEventListener('load', function () { Quickfile.onLoad(); }, false);
